
//Net Dance Server
//Server program for networked performance system One Bit Net Dance
//Andreja Andric
//2018

import java.net.*;
import java.io.*;
import java.util.*;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class NetDanceServer
{
//This is a list of, ehm, pointers to all the clients
    public Vector<NetDanceServerThread> clientList = new Vector<NetDanceServerThread>();
    public String[] OverallParamValues;
    public String[] OverallParams = {"MainPitch", "Duration", "Length", "DecayRate", "ScaleType", "PolyphonyType" };

    public String Melody;
    boolean listening;
    Lock l;
    int IdentifierCounter = 0;
    private int PortNumber;

    public NetDanceServer()
    {
       OverallParamValues = new String[] {"150", "128", "400", "4", "6", "1/1 1/2"};
       Melody = new String("0");
       try
       {
          FileReader  ul = new FileReader ("ServerConfiguration.txt");
          BufferedReader inFile = new BufferedReader(ul);
          String item;
          try
          {
             PortNumber = Integer.parseInt(inFile.readLine());
             System.out.println( "Port Number = "+PortNumber );
          
             inFile.close();
             ul.close ();
          }
          catch (NumberFormatException nfe)
          {
             System.out.println("Invalid file: ServerConfiguration.txt");
             return;
          }
          catch (IOException g)
          {
             System.out.println("Invalid file: ServerConfiguration.txt");
             return;
          }
       }
       catch (IOException g)
       {
          System.out.println("Couldn't read from file ServerConfiguration.txt");
          return;
       }

       try
       {
          Execute();
       }
       catch(IOException ioe)
       {
          System.out.println(ioe);
       }
    }

    void lockClientList()
    {
       l.lock();
    }

    void unlockClientList()
    {
       l.unlock();
    }

    public String GetMelody() { return Melody; }
  
    public int UpdateParams(String p)
    {
       int i;
       String[] params = p.split("\\s+");
       if (params.length<OverallParams.length)
          return -1;
       for(i=0;i<OverallParams.length;i++)
          OverallParamValues[i] = params[i];
       for(i=OverallParams.length;i<params.length;i++)
          OverallParamValues[OverallParams.length-1]+=" "+params[i];
       return 0;
    }
       
    public void UpdateMelody()
    {
       int i;
       int stop = 0;

       lockClientList();
       int[] pos = new int[clientList.size()];
       String[][] tones = new String[clientList.size()][];
       int maxlen = 0;
       for(i=0;i<clientList.size();i++)
       {
          pos[i] = 0;
          tones[i]=((NetDanceServerThread)clientList.elementAt(i)).GetMelody().split("\\W+");
          if(tones[i].length>maxlen)
             maxlen=tones[i].length;
       }
       StringBuilder sb = new StringBuilder();
       do
       {
          for(i=0;i<clientList.size();i++)
          {
             sb.append(" "+tones[i][pos[i]++]);
             if(pos[i]==tones[i].length)
                pos[i]=0;
          }
          stop++;
       }
       while(stop<maxlen);
       Melody = sb.toString().trim();
       unlockClientList();
    }

    public void remove(int i)
    {
       lockClientList();
       for(int j=0;j<clientList.size();j++)
       {
          if(clientList.elementAt(j).GetIndex()==i)
          {
             clientList.remove(j);
             break;
          }
       }
       System.out.println("There are " + clientList.size() + " clients left.");
       unlockClientList();
    }

    public void StopServerIfNoClients()
    {
       if(clientList.size()==0)
       {
          listening = false;
          //Now, since the last client has left we can stop the server
          //Since the server process is blocked at accepting socket connections
          //we'll make a fake one here to unblock it.
          try 
          {
            InetAddress NameAndIP = InetAddress.getLocalHost();
            Socket s = new Socket( NameAndIP.getHostAddress(), PortNumber);
          } catch (UnknownHostException e) {
            System.err.println("Don't know about host" );
            System.exit(1);
          } catch (IOException e) {
            System.err.println("Couldn't get I/O for the connection");
            System.exit(1);
          }
       }
    }

    public void Execute() throws IOException 
    {
        ServerSocket serverSocket = null;
        listening = true;
        System.out.println("Server started");

//temporary variable which serves for instantiating a new server thread
        NetDanceServerThread newClient;

//opens server socket on port 4444
        try {
            serverSocket = new ServerSocket(PortNumber);
        } catch (IOException e) {
            System.err.println("Could not listen on port: "+PortNumber);
            System.exit(-1);
        }

        l = new ReentrantLock();

        while (listening)
        {
//new server thread is created when a new client comes
           Socket s = serverSocket.accept();
           if(!listening)
              break;
           newClient = new NetDanceServerThread(s, this, IdentifierCounter++ );
//it is added to vector of threads; this vector is used by a server thread
//to discover 
           l.lock();
           try
           {
              clientList.addElement( newClient );
           }
           finally
           {
              l.unlock();
           }
           newClient.start();
           UpdateMelody();
           System.out.println("There are " + clientList.size() + " clients.");
        }
        System.out.println("Last client left. Exiting program. ");
        serverSocket.close();
    }

    public static void main(String[] args)
    {
       NetDanceServer ndns = new NetDanceServer();
    }
}
